import rclpy
from rclpy.node import Node
from rclpy.callback_groups import ReentrantCallbackGroup

from std_msgs.msg import Empty
from geometry_msgs.msg import PoseArray
from geometry_msgs.msg import Pose
from active_bo_msgs.msg import ActiveRLRequest
from active_bo_msgs.msg import ActiveRLResponse
from active_bo_msgs.msg import ActiveRLEvalRequest
from active_bo_msgs.msg import ActiveRLEvalResponse

from franka_iml_experiment.RepresentationModel.GaussianModelMultiDim import GaussianRBF

import numpy as np


class IML_Experiment(Node):

    def __init__(self):
        super().__init__('iml_node')

        rl_callback_group = ReentrantCallbackGroup()
        topic_callback_group = ReentrantCallbackGroup()

        self.traj_pub = self.create_publisher(PoseArray,
                                              'moveit_interface/task_space_trajectory',
                                              10,
                                              callback_group=rl_callback_group)

        self.exec_pub = self.create_publisher(Empty,
                                              'moveit_interface/execution',
                                              10)

        self.rl_subscription = self.create_subscription(ActiveRLRequest,
                                                        'active_rl_request',
                                                        self.rl_callback,
                                                        10,
                                                        callback_group=rl_callback_group)

        self.active_rl_pub = self.create_publisher(ActiveRLResponse,
                                                   'active_rl_response',
                                                   1,
                                                   callback_group=rl_callback_group)

        # interactive part
        self.eval_pub = self.create_publisher(ActiveRLEvalRequest,
                                              'active_rl_eval_request',
                                              1,
                                              callback_group=topic_callback_group)
        self.eval_sub = self.create_subscription(ActiveRLEvalResponse,
                                                 'active_rl_eval_response',
                                                 self.rl_eval_callback,
                                                 1,
                                                 callback_group=topic_callback_group)

        self.pub_rviz_model = self.create_publisher(Empty, 'moveit_interface/box', 10)
        msg = Empty()
        self.pub_rviz_model.publish(msg)

        # States
        self.eval_pending = False

        # Reward constants
        self.distance_pen = -0.01
        self.accel_pen = -0.01
        self.time_pen = -1.
        self.zone_pen = -20.
        self.goal_rew = 10.

        # Scene parameters
        self.square_center = np.array([0.55, 0.0])
        self.square_half_side = 0.1  # half of the side length
        self.circle_radius = 0.02

        # BO parameters
        self.start = {0: [0.55, -0.3]}
        self.end = [0.55, 0.3]
        self.nr_dims = None
        self.nr_weights = None
        self.nr_steps = None
        self.weight_preference = None
        self.weights = None
        self.policy = None

        # Policy parameter
        self.policy_model = None
        self.lower_bound = -1.0
        self.upper_bound = 1.0
        self.seed = None

    def rl_callback(self, msg):
        interactive_run = msg.interactive_run

        self.nr_weights = msg.nr_weights
        self.nr_steps = msg.nr_steps
        self.nr_dims = msg.nr_dims

        weight_dims = (self.nr_weights, self.nr_dims)

        self.weights = np.array(msg.weights, dtype=np.float64).reshape(weight_dims, order='F')

        self.policy_model = GaussianRBF(self.nr_steps, self.nr_weights, self.nr_dims,
                                        lower_bound=self.lower_bound, upper_bound=self.upper_bound, seed=self.seed,
                                        fixed_dims=self.start)

        self.policy_model.set_weights(np.around(self.weights, decimals=8))
        self.policy = self.policy_model.rollout()

        reward, not_perform, terminate_time = self.compute_reward(self.policy, self.nr_steps)

        short_policy = self.policy[:terminate_time, :]

        if interactive_run == 2:
            if not not_perform:
                pose_msg = PoseArray()
                for i in range(short_policy.shape[0]):
                    pose = Pose()
                    pose.position.x = short_policy[i, 0]
                    pose.position.y = short_policy[i, 1]
                    pose.position.z = 0.15
                    pose.orientation.x = 1.
                    pose.orientation.y = 0.
                    pose.orientation.z = 0.
                    pose.orientation.w = 0.
                    pose_msg.poses.append(pose)

                self.traj_pub.publish(pose_msg)
            else:
                self.get_logger().info(f'Trajectory is not valid!')

        if interactive_run == 1:
            if not not_perform:
                pose_msg = PoseArray()
                for i in range(short_policy.shape[0]):
                    pose = Pose()
                    pose.position.x = short_policy[i, 0]
                    pose.position.y = short_policy[i, 1]
                    pose.position.z = 0.15
                    pose.orientation.x = 1.
                    pose.orientation.y = 0.
                    pose.orientation.z = 0.
                    pose.orientation.w = 0.
                    pose_msg.poses.append(pose)

                # self.traj_pub.publish(pose_msg)
                # self.exec_pub.publish(Empty())

            rl_response = ActiveRLResponse()
            rl_response.weights = msg.weights
            rl_response.reward = reward
            rl_response.final_step = self.nr_steps
            if self.weight_preference is None:
                weight_preference = [False] * self.nr_weights * 2
            else:
                weight_preference = self.weight_preference

            rl_response.weight_preference = weight_preference

            self.active_rl_pub.publish(rl_response)

        else:
            self.get_logger().info('Interactive Learning started!')
            eval_msg = ActiveRLEvalRequest()
            eval_msg.weights = self.weights.flatten('F').tolist()
            eval_msg.policy = self.policy_model.rollout().flatten('F').tolist()
            eval_msg.nr_steps = self.nr_steps
            eval_msg.nr_weights = self.nr_weights

            self.eval_pub.publish(eval_msg)
            self.eval_pending = True

            pose_msg = PoseArray()
            for i in range(short_policy.shape[0]):
                pose = Pose()
                pose.position.x = short_policy[i, 0]
                pose.position.y = short_policy[i, 1]
                pose.position.z = 0.15
                pose.orientation.x = 1.
                pose.orientation.y = 0.
                pose.orientation.z = 0.
                pose.orientation.w = 0.
                pose_msg.poses.append(pose)

            self.traj_pub.publish(pose_msg)

    def rl_eval_callback(self, msg):
        if self.eval_pending:
            self.eval_pending = False
            weight_dims = (self.nr_weights, self.nr_dims)
            self.weights = np.array(msg.weights, dtype=np.float64).reshape(weight_dims, order='F')

            self.policy_model.set_weights(np.around(self.weights, decimals=8))
            self.policy = self.policy_model.rollout()

            self.weight_preference = msg.weight_preference

            self.get_logger().info('Active RL Eval: Responded!')

            reward, not_perform, terminate_time = self.compute_reward(self.policy, self.nr_steps)
            self.get_logger().info(f'Current Reward: {reward}')

            short_policy = self.policy[:terminate_time, :]

            if not not_perform:
                pose_msg = PoseArray()
                for i in range(short_policy.shape[0]):
                    pose = Pose()
                    pose.position.x = short_policy[i, 0]
                    pose.position.y = short_policy[i, 1]
                    pose.position.z = 0.15
                    pose.orientation.x = 1.
                    pose.orientation.y = 0.
                    pose.orientation.z = 0.
                    pose.orientation.w = 0.
                    pose_msg.poses.append(pose)

                self.traj_pub.publish(pose_msg)
                # self.exec_pub.publish(Empty())

            rl_response = ActiveRLResponse()
            rl_response.weights = msg.weights
            rl_response.reward = reward
            rl_response.final_step = self.nr_steps
            if self.weight_preference is None:
                weight_preference = [False] * self.nr_weights * 2
            else:
                weight_preference = self.weight_preference

            rl_response.weight_preference = weight_preference

            self.active_rl_pub.publish(rl_response)

    def compute_reward(self, y_track, time):
        # 1. Check if any (x,y) position is within the square
        within_square = np.any(
            (y_track[:, 0] > self.square_center[0] - self.square_half_side) &
            (y_track[:, 0] < self.square_center[0] + self.square_half_side) &
            (y_track[:, 1] > self.square_center[1] - self.square_half_side) &
            (y_track[:, 1] < self.square_center[1] + self.square_half_side)
        )

        # If the robot moves into the forbidden zone it won't perform trajectory and get penalized
        if within_square:
            return self.zone_pen + time * self.time_pen, within_square, 0

        # 2. Check if any (x,y) position is within the circle
        t = 0
        for i in range(y_track.shape[0]):
            distance_to_circle_center = np.linalg.norm(y_track[i, :] - self.end)
            if distance_to_circle_center < self.circle_radius:
                t = i
                break

        if t == 0:
            t = y_track.shape[0]

        y_track = y_track[:t, :]

        penalty = t * self.time_pen

        # 3. Compute the Euclidean distance from each (x,y) position to the center of the square
        distances_to_square_center = np.linalg.norm(y_track - self.square_center, axis=1)
        total_distance_norm = np.sum(distances_to_square_center)

        penalty += total_distance_norm * self.distance_pen

        return penalty, within_square, t


def main(args=None):
    rclpy.init(args=args)

    dmp_node = IML_Experiment()

    rclpy.spin(dmp_node)

    dmp_node.destroy_node()
    rclpy.shutdown()


if __name__ == '__main__':
    main()
